import numpy as np
import numpy.linalg as la
##########################################################################################
### Problem Physical Parameters:
def getE():
    return 1
def getNu():
    return 0.3
##########################################################################################
### Neumann Boundary Conditions and vonMises
def neumann_bcs(z,nz,side):
    t = h(z,side) * nz
    return 1j*np.imag(t)
def vonMises(sxx,syy,sxy):
    # Return vonMises effective Stress for a plane strain problem out of a Stress Tensor:
    nu = getNu()
    return np.sqrt((1-nu*(1-nu))*(sxx+syy)**2 - 3*(sxx*syy-sxy**2))
##########################################################################################
### Auxiliary Functions for Complex/Real Operators
def Op_to_real(Op):
    OpR = np.vstack([np.hstack([np.real(Op),-np.imag(Op)]),np.hstack([np.imag(Op),np.real(Op)])])
    OpRbar = np.vstack([np.hstack([np.real(Op),np.imag(Op)]),np.hstack([np.imag(Op),-np.real(Op)])])
    return OpR, OpRbar
def Op_to_complex(Op):
    m, n = np.shape(Op)
    Opr = Op[:int(m/2),:int(n/2)]
    Opi = Op[int(m/2):,:int(n/2)]
    return Opr + 1j*Opi
def vec_to_real(v):
    return np.concatenate((v.real,v.imag))
def vec_to_complex(v):
    vr = v[:int(np.size(v)/2)]
    vi = v[int(np.size(v)/2):]
    return vr + 1j*vi
def vander_op(t):
    n = np.size(t)
    vandermat = np.vander(t,increasing=True)
    vanderdiff = np.repeat(np.arange(n).reshape(1,-1),n,axis=0) * (vandermat / np.repeat(t.reshape(-1,1),n,axis=1))
    diffmat = vanderdiff @ la.inv(vandermat)
    return vandermat, vanderdiff, diffmat
##########################################################################################
### Definition of Operators:
def h(z,side):
    sign = (np.vstack([(abs(np.real(z)) >= side/2 - 1e-12),(abs(np.imag(z)) >= side/2 - 1e-12)])).any(axis=0)*1
    return sign
def rho(z,bp,tp,side):
    rho = (np.sqrt(z-bp)/np.sqrt(z-tp))
    sign = 2*(np.vstack([(abs(np.real(z)) >= side/2 - 1e-12),(abs(np.imag(z)) >= side/2 - 1e-12)])).any(axis=0) - 1
    return sign*rho
def g(z,nz,tr,w,bp,tp):
    tauz = z.reshape(1,-1) - z.reshape(-1,1)
    nbarn = (np.conj(nz)/nz).reshape(-1,1)
    nbartr = np.conj(nz) * tr
    G1 = (1/(2j * np.pi)) * 1/np.conj(tauz) * np.conj(w).reshape(1,-1)
    np.fill_diagonal(G1,0)
    g1 = nbarn.reshape(-1,1) * (np.sum((nz*tr).reshape(1,-1)*G1,axis=1)).reshape(-1,1)
    g2 = -nbartr.reshape(-1,1) * (np.sum(G1,axis=1)).reshape(-1,1)
    g3 = nbartr.reshape(-1,1)/(2j * np.pi)*(np.conj(np.log(bp - z.reshape(-1,1))) - np.conj(np.log(tp - z.reshape(-1,1))))
    return (g1+g2+g3).reshape(-1)
def Q(z,w):
    Q = (z.reshape(-1,1)*0+1)/(1j*np.pi) * w.reshape(1,-1)
    return Q
def iP_0(z,w,side,A):
    z_bdr = h(z,side) * z
    # Part that acts on the function:
    iP_0_1 = -1j*(z.reshape(-1,1)*0+1)/(4*A) * np.conj(z_bdr).reshape(1,-1) * w.reshape(1,-1)
    # Part that acts on the conjugate of the function:
    iP_0_2 = -1j*(z.reshape(-1,1)*0+1)/(4*A) * z_bdr.reshape(1,-1) * np.conj(w).reshape(1,-1)
    return iP_0_1, iP_0_2
def M1_rho_inv(z,nz,w,K,npanels,npts):
    tauz = z.reshape(1,-1) - z.reshape(-1,1)
    M1_rho_inv = 1/(1j*np.pi) * 1/tauz * w.reshape(1,-1)
    np.fill_diagonal(M1_rho_inv,0)
    c = 1 - np.sum(M1_rho_inv,axis=1)
    diagon = np.arange(M1_rho_inv.shape[0])
    M1_rho_inv[diagon,diagon] =  c
    for i in range(0,npanels):
        M1_rho_inv[i*npts:(i+1)*npts,i*npts:(i+1)*npts] += 1/(1j*np.pi) * w[0,i*npts:(i+1)*npts].reshape(1,-1) * K[i,:,:]
    return M1_rho_inv
def M1_rho_inv_loc(zl,nzl,wl,zg,nzg,wg,K,npts):
    N = int(zl.shape[0]/npts)
    tauzl = zl.reshape(1,-1) - zl.reshape(-1,1)
    tauzgl = zg.reshape(1,-1) - zl.reshape(-1,1)
    M1_rho_inv = 1/(1j*np.pi) * 1/tauzl * wl.reshape(1,-1)
    np.fill_diagonal(M1_rho_inv,0)
    M = 1/(1j*np.pi) * 1/tauzgl * wg.reshape(1,-1)
    c = 1 - np.sum(M1_rho_inv,axis=1) - np.sum(M,axis=1)
    diagon = np.arange(M1_rho_inv.shape[0])
    M1_rho_inv[diagon,diagon] =  c
    for i in range(0,N):
        M1_rho_inv[i*npts:(i+1)*npts,i*npts:(i+1)*npts] += 1/(1j*np.pi) * wl[0,i*npts:(i+1)*npts].reshape(1,-1) * K[i,:,:]
    return M1_rho_inv
def M3_rho(z,nz,w):
    tauz = z.reshape(1,-1) - z.reshape(-1,1)
    nbarn = (np.conj(nz)/nz).reshape(-1,1)
    # Part that acts on the function:
    M31 = 1/(2j*np.pi) * (1/tauz + nbarn/np.conj(tauz)) * w.reshape(1,-1)
    np.fill_diagonal(M31,0) # Second Derivative is zero
    # Part that acts on the conjugate of the function:
    M32 = 1/(2j*np.pi) * (1/np.conj(tauz) + nbarn*tauz/(np.conj(tauz))**2) * np.conj(w).reshape(1,-1)
    np.fill_diagonal(M32,0) # Second Derivative is zero
    return M31,M32
##########################################################################################
### Definition of Operators to obtain stresses out of density:
def PHI(zs,zt,sigma,w):
    tauz = zs.reshape(1,-1) - zt.reshape(-1,1)
    PHI = 1/(2j*np.pi) * 1/tauz * sigma.reshape(1,-1) * w.reshape(1,-1)
    return PHI.sum(axis=1)
def PSI(zs,nzs,tr,zt,sigma,w):
    tauz = zs.reshape(1,-1) - zt.reshape(-1,1)
    PSI = -1/(2j*np.pi) * ( 1/tauz * np.conj(sigma).reshape(1,-1) * np.conj(w).reshape(1,-1) 
    + 1/(tauz)**2 * np.conj(zs).reshape(1,-1) *  sigma.reshape(1,-1) * w.reshape(1,-1)
    + 1/tauz * np.conj(nzs).reshape(1,-1) * np.conj(tr).reshape(1,-1) * w.reshape(1,-1) )
    return PSI.sum(axis=1)
def PHIPR(zs,zt,sigma,w):
    tauz = zs.reshape(1,-1) - zt.reshape(-1,1)
    PHIPR = 1/(2j*np.pi) * 1/(tauz)**2 * sigma.reshape(1,-1) * w.reshape(1,-1)
    return PHIPR.sum(axis=1)