import numpy as np
import scipy.special
import matplotlib.pyplot as plt
from operators import *
##########################################################################################
class GeometricInfo:
    def __init__(self, side):
        self.side = side
        self.area = side**2
        self.intersect = -side/2+0j # Branch Point Coord
        self.tip = 0+0j # Tip Point Coord
        self.crn_1 = side/2*(-1 - 1j)
        self.crn_2 = side/2*(1 - 1j)
        self.crn_3 = side/2*(1 + 1j)
        self.crn_4 = side/2*(-1 + 1j)
        plt.plot(
            [self.tip.real,self.intersect.real,self.crn_1.real,self.crn_2.real,self.crn_3.real,self.crn_4.real,self.intersect.real],
            [self.tip.imag,self.intersect.imag,self.crn_1.imag,self.crn_2.imag,self.crn_3.imag,self.crn_4.imag,self.intersect.imag],
            "-",linewidth=1,color='#B01B1B')
        plt.show()
class QuadratureInfo:
    def __init__(self, npanels, npts,geom_info):
        self.nintervals = npanels # Total number of intervals (Crack+Boundary):
        self.nintervals_crk = int(npanels/9) # Number of intervals corresponding to the crack line:
        # Panel/Node indices of intersection and corners:
        self.idx_intersect = int(npanels/9)
        self.idx_crn_1 = 2*int(npanels/9)
        self.idx_crn_2 = 4*int(npanels/9)
        self.idx_crn_3 = 6*int(npanels/9)
        self.idx_crn_4 = 8*int(npanels/9)
        par_length = geom_info.side*9/2
        intervals = np.linspace(0, par_length, npanels+1)
        self.npoints = npts
        self.shape = (self.nintervals, self.npoints)
        #################################################
        # Gauss-Legendre Quadrature:
        ref_info_gl = scipy.special.legendre(self.npoints).weights
        ref_nodes_gl = ref_info_gl[:,0]
        ref_weights_gl = ref_info_gl[:,2]
        # Gauss-Jacobi Quadrature:
        ref_info_sp = scipy.special.jacobi(self.npoints,0.0,-0.5).weights
        ref_nodes_sp = ref_info_sp[:,0]
        ref_weights_sp = ref_info_sp[:,2]
        #################################################
        self.ref_nodes = np.repeat(ref_nodes_gl.reshape(1,-1),npanels,axis=0)
        self.ref_weights = np.repeat(ref_weights_gl.reshape(1,-1),npanels,axis=0)
        #################################################
        # Setup special quadrature rules for panels adjacent to crack tip:
        # Crack Tip:
        self.ref_nodes[0,:] = ref_nodes_sp
        self.ref_weights[0,:] = ref_weights_sp
        #################################################
        par_intv_length = intervals[1] - intervals[0]
        self.par_nodes = np.zeros(self.shape)
        self.vandermat = np.zeros((npanels, self.npoints, self.npoints))
        self.vanderdiff = np.zeros((npanels, self.npoints, self.npoints))
        self.diffmat = np.zeros((npanels, self.npoints, self.npoints))
        for i in range(npanels):
            a, b = intervals[i:i+2]
            assert abs((b-a) - par_intv_length) < 1e-10
            self.par_nodes[i] = self.ref_nodes[i,:]*par_intv_length*0.5 + (intervals[0]+intervals[1])*0.5
            self.vandermat[i,:,:], self.vanderdiff[i,:,:], self.diffmat[i,:,:] = vander_op(self.ref_nodes[i,:])
        cdir = (np.array([-1,-1j,1,1,1j,1j,-1,-1,-1j]).repeat(self.nintervals_crk)).reshape(npanels,1)
        self.panel_nodes = np.hstack([0,np.cumsum(cdir*par_intv_length)[0:-1]])
        self.curve_nodes = (cdir*self.par_nodes + self.panel_nodes.reshape(-1,1)).reshape(-1)
        self.curve_deriv = (cdir*np.ones(self.shape)).reshape(-1)
        self.curve_speed = abs(self.curve_deriv)
        tangent = self.curve_deriv / self.curve_speed
        self.normals = np.imag(tangent) - 1j * np.real(tangent)
        self.panel_lengths = np.sum(self.curve_speed.reshape(npanels, -1) * self.ref_weights * par_intv_length / 2, 1)
        #################################################
        self.curve_weights_plain = (tangent.reshape(npanels, -1) * self.ref_weights * par_intv_length / 2).reshape(-1)
        self.curve_weights_rho_inv = self.curve_weights_plain / rho(self.curve_nodes,geom_info.intersect,geom_info.tip,geom_info.side)
        self.curve_weights_rho = self.curve_weights_plain * rho(self.curve_nodes,geom_info.intersect,geom_info.tip,geom_info.side)
def refine_tip(nsub,quad_info,geom_info):
    npts = quad_info.shape[1]
    zt = geom_info.tip
    z1 = np.copy(quad_info.panel_nodes[1])
    zfr = quad_info.curve_nodes[0:npts]
    nzfr = quad_info.normals[0:npts]
    wfr_s = quad_info.curve_weights_plain[0:npts]
    wfr_r = quad_info.curve_weights_plain[npts:2*npts]
    Ktemp_s = quad_info.diffmat[0,:,:]
    Ktemp_r = quad_info.diffmat[1,:,:]
    zloc = 1j * np.arange(npts)
    nzloc = 1j * np.arange(npts)
    wloc = 1j * np.arange(npts)
    Kloc = np.zeros((1,npts,npts))
    for i in range(nsub):
        zloc = zloc[npts:]
        zloc = np.insert(zloc,0,np.concatenate((zt + (zfr - zt)/2,z1 + (zfr - z1)/2)))
        nzloc = nzloc[npts:]
        nzloc = np.append(nzloc,np.concatenate((nzfr,nzfr)))
        wloc = wloc[npts:]
        wloc = np.insert(wloc,0,np.concatenate((wfr_s,wfr_r))/2**(i+1))
        Kloc = Kloc[npts:,:,:]
        Kloc = np.insert(Kloc,0,np.concatenate((Ktemp_s.reshape(1,npts,npts),Ktemp_r.reshape(1,npts,npts)),axis=0),axis=0)
        zfr = zloc[0:npts]
        z1 = (zt + z1) /2
    rholoc = rho(zloc,geom_info.intersect,geom_info.tip,geom_info.side)
    return zloc, nzloc, wloc, rholoc, Kloc
def refine_crn(crn_idx,nsub,quad_info,geom_info):
    npts = quad_info.shape[1]
    zc = np.copy(quad_info.panel_nodes[crn_idx])
    zl = np.copy(quad_info.panel_nodes[crn_idx-1])
    zr = np.copy(quad_info.panel_nodes[crn_idx+1])
    zfr_l = quad_info.curve_nodes[(crn_idx-1)*npts:crn_idx*npts]
    zfr_r = quad_info.curve_nodes[crn_idx*npts:(crn_idx+1)*npts]
    nzfr_l = quad_info.normals[(crn_idx-1)*npts:crn_idx*npts]
    nzfr_r = quad_info.normals[crn_idx*npts:(crn_idx+1)*npts]
    wfr_l = quad_info.curve_weights_plain[(crn_idx-1)*npts:crn_idx*npts]
    wfr_r = quad_info.curve_weights_plain[crn_idx*npts:(crn_idx+1)*npts]
    Ktemp_l = quad_info.diffmat[crn_idx-1,:,:]
    Ktemp_r = quad_info.diffmat[crn_idx,:,:]
    zloc = 1j * np.arange(2*npts)
    nzloc = 1j * np.arange(2*npts)
    wloc = 1j * np.arange(2*npts)
    Kloc = np.zeros((2,npts,npts))
    for i in range(nsub):
        zloc = np.delete(zloc,np.arange(int(zloc.size/2)-npts,int(zloc.size/2)+npts))
        zloc = np.insert(zloc,int(zloc.size/2),np.concatenate((zl + (zfr_l - zl)/2,zc + (zfr_l - zc)/2,zc + (zfr_r - zc)/2,zr + (zfr_r- zr)/2)))
        nzloc = np.delete(nzloc,np.arange(int(nzloc.size/2)-npts,int(nzloc.size/2)+npts))
        nzloc = np.insert(nzloc,int(nzloc.size/2),np.concatenate((nzfr_l,nzfr_l,nzfr_r,nzfr_r)))
        wloc = np.delete(wloc,np.arange(int(wloc.size/2)-npts,int(wloc.size/2)+npts))
        wloc = np.insert(wloc,int(wloc.size/2),np.concatenate((wfr_l,wfr_l,wfr_r,wfr_r))/2**(i+1))
        Kloc = np.delete(Kloc,np.arange(int(Kloc.shape[0]/2)-1,int(Kloc.shape[0]/2)+1),axis=0)
        Kloc = np.insert(Kloc,int(Kloc.shape[0]/2),np.concatenate((Ktemp_l.reshape(1,npts,npts),Ktemp_l.reshape(1,npts,npts),Ktemp_r.reshape(1,npts,npts),Ktemp_r.reshape(1,npts,npts)),axis=0),axis=0)
        zfr_l = zloc[int(zloc.size/2)-npts:int(zloc.size/2)]
        zfr_r = zloc[int(zloc.size/2):int(zloc.size/2)+npts]
        zl = (zl + zc) /2
        zr = (zr + zc) /2
    rholoc = rho(zloc,geom_info.intersect,geom_info.tip,geom_info.side)
    return zloc, nzloc, wloc, rholoc, Kloc
def refine_int(nsub,quad_info,geom_info):
    npts = quad_info.shape[1]
    int_idx = quad_info.idx_intersect
    zc = np.copy(quad_info.panel_nodes[int_idx])
    zl = np.copy(quad_info.panel_nodes[int_idx-1])
    zd = np.copy(quad_info.panel_nodes[int_idx+1])
    zu = np.copy(quad_info.panel_nodes[-1])
    zfr_l = quad_info.curve_nodes[(int_idx-1)*npts:int_idx*npts]
    zfr_d = quad_info.curve_nodes[int_idx*npts:(int_idx+1)*npts]
    zfr_u = quad_info.curve_nodes[-npts:]
    nzfr_l = quad_info.normals[(int_idx-1)*npts:int_idx*npts]
    nzfr_d = quad_info.normals[int_idx*npts:(int_idx+1)*npts]
    nzfr_u = quad_info.normals[-npts:]
    wfr_l_s = quad_info.curve_weights_plain[(int_idx-1)*npts:int_idx*npts]
    wfr_d_s = quad_info.curve_weights_plain[int_idx*npts:(int_idx+1)*npts]
    wfr_u_s = quad_info.curve_weights_plain[-npts:]
    wfr_l_r = quad_info.curve_weights_plain[(int_idx-2)*npts:(int_idx-1)*npts]
    wfr_d_r = quad_info.curve_weights_plain[(int_idx+1)*npts:(int_idx+2)*npts]
    wfr_u_r = quad_info.curve_weights_plain[-2*npts:-npts]
    Ktemp_l_s = quad_info.diffmat[int_idx-1,:,:]
    Ktemp_d_s = quad_info.diffmat[int_idx,:,:]
    Ktemp_u_s = quad_info.diffmat[-1,:,:]
    Ktemp_l_r = quad_info.diffmat[int_idx-2,:,:]
    Ktemp_d_r = quad_info.diffmat[int_idx+1,:,:]
    Ktemp_u_r = quad_info.diffmat[-2,:,:]
    zloc = 1j * np.arange(2*npts)
    nzloc = 1j * np.arange(2*npts)
    wloc = 1j * np.arange(2*npts)
    zloc1 = 1j * np.arange(npts)
    nzloc1 = 1j * np.arange(npts)
    wloc1 = 1j * np.arange(npts)
    Kloc = np.zeros((2,npts,npts))
    Kloc1 = np.zeros((1,npts,npts))
    for i in range(nsub):
        zloc = np.delete(zloc,np.arange(int(zloc.size/2)-npts,int(zloc.size/2)+npts))
        zloc = np.insert(zloc,int(zloc.size/2),np.concatenate((zl + (zfr_l - zl)/2,zc + (zfr_l - zc)/2,zc + (zfr_d - zc)/2,zd + (zfr_d- zd)/2)))
        nzloc = np.delete(nzloc,np.arange(int(nzloc.size/2)-npts,int(nzloc.size/2)+npts))
        nzloc = np.insert(nzloc,int(nzloc.size/2),np.concatenate((nzfr_l,nzfr_l,nzfr_d,nzfr_d)))
        wloc = np.delete(wloc,np.arange(int(wloc.size/2)-npts,int(wloc.size/2)+npts))
        wloc = np.insert(wloc,int(wloc.size/2),np.concatenate((wfr_l_r,wfr_l_s,wfr_d_s,wfr_d_r))/2**(i+1))
        Kloc = np.delete(Kloc,np.arange(int(Kloc.shape[0]/2)-1,int(Kloc.shape[0]/2)+1),axis=0)
        Kloc = np.insert(Kloc,int(Kloc.shape[0]/2),np.concatenate((Ktemp_l_r.reshape(1,npts,npts),Ktemp_l_s.reshape(1,npts,npts),Ktemp_d_s.reshape(1,npts,npts),Ktemp_d_r.reshape(1,npts,npts)),axis=0),axis=0)
        zfr_l = zloc[int(zloc.size/2)-npts:int(zloc.size/2)]
        zfr_d = zloc[int(zloc.size/2):int(zloc.size/2)+npts]
        zl = (zl + zc) /2
        zd = (zd + zc) /2
        zloc1 = zloc1[0:-npts]
        zloc1 = np.append(zloc1,np.concatenate((zu + (zfr_u - zu)/2,zc + (zfr_u - zc)/2)))
        nzloc1 = nzloc1[0:-npts]
        nzloc1 = np.append(nzloc1,np.concatenate((nzfr_u,nzfr_u)))
        wloc1 = wloc1[0:-npts]
        wloc1 = np.append(wloc1,np.concatenate((wfr_u_r,wfr_u_s))/2**(i+1))
        Kloc1 = Kloc1[0:-1,:,:]
        Kloc1 = np.append(Kloc1,np.concatenate((Ktemp_u_r.reshape(1,npts,npts),Ktemp_u_s.reshape(1,npts,npts)),axis=0),axis=0)
        zfr_u = zloc1[-npts:]
        zu = (zu + zc) /2
    rholoc = rho(zloc,geom_info.intersect,geom_info.tip,geom_info.side)
    rholoc1 = rho(zloc1,geom_info.intersect,geom_info.tip,geom_info.side)
    return zloc, nzloc, wloc, rholoc, Kloc, zloc1, nzloc1, wloc1, rholoc1, Kloc1